@@ -182,6 +182,8 @@ const (
 	// Error during encoding or when writing to the output
 	ErrorEncoding = C.heif_error_Encoding_error
 
+	ErrorEndOfSequence = C.heif_error_End_of_sequence
+
 	// Application has asked for a color profile type that does not exist
 	ErrorColorProfileDoesNotExist = C.heif_error_Color_profile_does_not_exist
 
@@ -211,6 +213,8 @@ const (
 
 	SuberrorNoMetaBox = C.heif_suberror_No_meta_box
 
+	SuberrorNoMoovBox = C.heif_suberror_No_moov_box
+
 	SuberrorNoHdlrBox = C.heif_suberror_No_hdlr_box
 
 	SuberrorNoHvcCBox = C.heif_suberror_No_hvcC_box
@@ -365,6 +369,8 @@ const (
 
 	SuberrorUnsupportedHeaderCompressionMethod = C.heif_suberror_Unsupported_header_compression_method
 
+	SubErrorUnsupportedTrackType = C.heif_suberror_Unsupported_track_type
+
 	// --- Encoder_plugin_error ---
 
 	SuberrorUnsupportedBitDepth = C.heif_suberror_Unsupported_bit_depth
